<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:env="http://schemas.xmlsoap.org/soap/envelope/" xmlns:dp="http://www.datapower.com/extensions" extension-element-prefixes="dp" xmlns="">
	<!-- -->
	<xsl:template match="/">
		<!-- -->
		<!-- Fetch the error information captured by On-Error rule -->
		<!-- Also capture the x-dp-response-code in case this rule has -->
		<!-- been executed by a backend failure -->
		<!-- -->
		<xsl:variable name="errorResponse" select="dp:variable('var://context/errorContext/errorResponse')"/>
		<xsl:variable name="errorReason" select="dp:variable('var://context/errorContext/errorReason')"/>
		<xsl:variable name="x-dp-response-code" select="normalize-space(dp:response-header('x-dp-response-code'))"/>
		<!-- -->
		<xsl:message>
			<xsl:value-of select="concat('errorResponse=', $errorResponse)"/>
			<xsl:value-of select="concat(', errorReason=', $errorReason)"/>
			<xsl:value-of select="concat(', x-dp-response-code=', $x-dp-response-code)"/>
		</xsl:message>
		<!-- -->
		<!-- When x-dp-response-code is set, a DP Service on the backend has failed -->
		<!-- If so, use it's code, else get the code set by the error rule -->
		<!-- -->
		<xsl:variable name="protocolResponse">
			<xsl:choose>
				<xsl:when test="$x-dp-response-code!=''">
					<xsl:value-of select="substring($x-dp-response-code, 1, 3)"/>
				</xsl:when>
				<xsl:otherwise>
					<xsl:value-of select="$errorResponse"/>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:variable>
		<!-- -->
		<!-- Set the services Protocol Reponse -->
		<!-- -->
		<dp:set-variable name="'var://service/error-protocol-response'" value="$protocolResponse"/>
		<xsl:message>
			<xsl:value-of select="concat('var://service/error-protocol-response set to ', $protocolResponse)"/>
		</xsl:message>
		<!-- -->
		<!-- Get information about the transaction -->
		<!-- -->
		<xsl:variable name="transaction-id" select="dp:variable('var://service/transaction-id')"/>
		<xsl:variable name="transaction-client" select="dp:variable('var://service/transaction-client')"/>
		<xsl:variable name="URL-in" select="dp:variable('var://service/URL-in')"/>
		<xsl:variable name="processor-name" select="dp:variable('var://service/processor-name')"/>
		<xsl:variable name="processor-type" select="dp:variable('var://service/processor-type')"/>
		<!-- -->
		<!-- Produce a soap fault with all the aquired information -->
		<!-- -->
		<env:Fault>
			<faultcode>
				<xsl:value-of select="$protocolResponse"/>
			</faultcode>
			<faultstring>
				<xsl:choose>
					<xsl:when test="$x-dp-response-code!=''">
						<xsl:value-of select="substring($error-protocol-response, 5)"/>
					</xsl:when>
					<xsl:otherwise>
						<xsl:value-of select="$errorReason"/>
					</xsl:otherwise>
				</xsl:choose>
			</faultstring>
			<detail>
				<transactionInfo>
					<transaction-id>
						<xsl:value-of select="$transaction-id"/>
					</transaction-id>
					<transaction-client>
						<xsl:value-of select="$transaction-client"/>
					</transaction-client>
					<URL-in>
						<xsl:value-of select="$URL-in"/>
					</URL-in>
					<processor-name>
						<xsl:value-of select="$processor-name"/>
					</processor-name>
					<processor-type>
						<xsl:value-of select="$processor-type"/>
					</processor-type>
				</transactionInfo>
				<!-- -->
				<!-- Get information about the error which was captured in On-Error rule -->
				<!-- -->
				<xsl:variable name="errorInfo" select="dp:variable('var://context/errorContext/errorInfo')"/>
				<xsl:copy-of select="$errorInfo"/>
			</detail>
		</env:Fault>
	</xsl:template>
	<!-- -->
</xsl:stylesheet>
